/*++
/* NAME
/*	read_wait 3
/* SUMMARY
/*	wait until descriptor becomes readable
/* SYNOPSIS
/*	#include <iostuff.h>
/*
/*	int	read_wait(fd, timeout)
/*	int	fd;
/*	int	timeout;
/* DESCRIPTION
/*	read_wait() blocks the current process until the specified file
/*	descriptor becomes readable, or until the deadline is exceeded.
/*
/*	Arguments:
/* .IP fd
/*	File descriptor in the range 0..FD_SETSIZE.
/* .IP timeout
/*	If positive, deadline in seconds. A zero value effects a poll.
/*	A negative value means wait until something happens.
/* DIAGNOSTICS
/*	Panic: interface violation. All system call errors are fatal.
/*
/*	A zero result means success.  When the specified deadline is
/*	exceeded, read_wait() returns -1 and sets errno to ETIMEDOUT.
/* LICENSE
/* .ad
/* .fi
/*	The Secure Mailer license must be distributed with this software.
/* AUTHOR(S)
/*	Wietse Venema
/*	IBM T.J. Watson Research
/*	P.O. Box 704
/*	Yorktown Heights, NY 10598, USA
/*--*/

/* System library. */

#include <sys_defs.h>
#include <sys/time.h>
#include <signal.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>

#ifdef USE_SYSV_POLL
#include <poll.h>
#endif

#if defined(USE_SELECT_POLL)
#error "don't define USE_SELECT_POLL"
#endif

#if !defined(USE_SYSV_POLL)
#define USE_SELECT_POLL
#endif

#ifdef USE_SYS_SELECT_H
#include <sys/select.h>
#endif

#ifdef USE_KQUEUE_POLL
#include <sys/event.h>

int     readwrite_kq = -1;		/* shared, NOT public */

#endif

/* Utility library. */

#include <msg.h>
#include <iostuff.h>

/* read_wait - block with timeout until file descriptor is readable */

int     read_wait(int fd, int timeout)
{
    const char *myname = "read_wait";

#ifdef USE_SELECT_POLL
    fd_set  read_fds;
    fd_set  except_fds;
    struct timeval tv;
    struct timeval *tp;

    /*
     * Sanity checks.
     */
    if (FD_SETSIZE <= fd)
	msg_panic("%s: descriptor %d does not fit FD_SETSIZE %d",
		  myname, fd, FD_SETSIZE);

    /*
     * Use select() so we do not depend on alarm() and on signal() handlers.
     * Restart the select when interrupted by some signal. Some select()
     * implementations reduce the time to wait when interrupted, which is
     * exactly what we want.
     */
    FD_ZERO(&read_fds);
    FD_SET(fd, &read_fds);
    FD_ZERO(&except_fds);
    FD_SET(fd, &except_fds);
    if (timeout >= 0) {
	tv.tv_usec = 0;
	tv.tv_sec = timeout;
	tp = &tv;
    } else {
	tp = 0;
    }

    for (;;) {
	switch (select(fd + 1, &read_fds, (fd_set *) 0, &except_fds, tp)) {
	case -1:
	    if (errno != EINTR)
		msg_fatal("select: %m");
	    continue;
	case 0:
	    errno = ETIMEDOUT;
	    return (-1);
	default:
	    return (0);
	}
    }
#endif

    /*
     * System-V poll() is optimal for polling a few descriptors.
     */
#ifdef USE_SYSV_POLL
    struct pollfd pollfd;

#define WAIT_FOR_EVENT	(-1)

    pollfd.fd = fd;
    pollfd.events = POLLIN;
    for (;;) {
	switch (poll(&pollfd, 1, timeout < 0 ?
		     WAIT_FOR_EVENT : timeout * 1000)) {
	case -1:
	    if (errno != EINTR)
		msg_fatal("poll: %m");
	    continue;
	case 0:
	    errno = ETIMEDOUT;
	    return (-1);
	default:
	    if (pollfd.revents & POLLNVAL)
		msg_fatal("poll: %m");
	    return (0);
	}
    }
#endif

    /*
     * The kqueue()/kevent() mechanism is optimal for polling huge amounts of
     * descriptors. The following code is for educational purposes only. Do
     * not use it for serious work - it makes too many system calls.
     * 
     * Note to readers with short attention spans: I wrote "do not use the
     * following code". I didn't write "do not use kqueue()/kevent().
     */
#if 0
    struct timespec ts;
    struct timespec *tp;
    struct kevent rreq;
    struct kevent revnt;

    if (readwrite_kq < 0 && (readwrite_kq = kqueue()) == -1)
	msg_fatal("kqueue: %m");

    if (timeout >= 0) {
	ts.tv_nsec = 0;
	ts.tv_sec = timeout;
	tp = &ts;
    } else {
	tp = 0;
    }

#define RETURN(x) do { \
	EV_SET(&rreq, fd, EVFILT_READ, EV_DELETE, 0, 0, 0); \
	if (kevent(readwrite_kq, &rreq, 1, 0, 0, 0) < 0) \
	    msg_fatal("kevent: %m"); \
	return (x); \
    } while (0)

    EV_SET(&rreq, fd, EVFILT_READ, EV_ADD, 0, 0, 0);
    if (kevent(readwrite_kq, &rreq, 1, 0, 0, 0) < 0)
	msg_fatal("kevent: %m");
    for (;;) {
	switch (kevent(readwrite_kq, 0, 0, &revnt, 1, tp)) {
	case -1:
	    if (errno != EINTR)
		msg_fatal("kevent: %m");
	    continue;
	case 0:
	    errno = ETIMEDOUT;
	    RETURN(-1);
	default:
	    if (revnt.ident != fd)
		msg_panic("%s: event for unexpected descriptor: %d != %d",
			  myname, revnt.ident, fd);
	    RETURN(0);
	}
    }
#endif
}
